"""Caching mechanisms
"""
from builtins import str
from builtins import object

from os import path, remove
import requests
import datetime
import time

from sqlalchemy import Column, String, Integer, DateTime
from sqlalchemy.orm.exc import NoResultFound
# noinspection PyPackageRequirements
from lxml import html

from ozelot import config, client
from ozelot.orm import base


# a singleton cache instance
_request_cache = None


# custom error codes for request errors
ERROR_CONNECTION_ERROR = 999
ERROR_XPATH_NOT_FOUND = 998


def reset_request_cache():
    """Clear singleton :class:`RequestCache` instance
    """
    global _request_cache

    if _request_cache is not None:
        _request_cache.close()
        _request_cache.clear()
        _request_cache = None


def get_request_cache():
    """Get a singleton :class:`RequestCache` instance
    """
    global _request_cache

    if _request_cache is None:
        _request_cache = RequestCache()

    return _request_cache


class CachedRequest(base.Base):
    """ORM object class for cached URL request data
    """

    url = Column(String(2048), index=True)
    xpath = Column(String(512), index=True)
    content = Column(String)
    status_code = Column(Integer)
    queried_on = Column(DateTime)


class RequestCache(object):
    """A simple cache for (HTTP) requests.

    Uses an SQLite DB (and a simple ORM) to store cached request results.

    .. warning:: This implementation is NOT thread-safe!
    """

    STATUS_CODE_SUFFIX = '___status_code'

    def __init__(self):
        """Initialize the cache
        """
        self.db_path = config.REQUEST_CACHE_PATH

        connection_string = "sqlite:///" + self.db_path

        # timestamp of last query
        self.last_query = None

        if path.exists(self.db_path):
            # if file exists, assume that the DB is initialized
            config.logger.debug("Opening an existing request cache file " + self.db_path)
            self.client = client.Client(connection_string=connection_string)
        else:
            # create and initialize a new request cache DB
            config.logger.debug("Creating a new request cache file " + self.db_path)
            self.client = client.Client(connection_string=connection_string)
            CachedRequest().create_table(self.client)

        self.session = self.client.create_session()

    def __del__(self):
        """Clean up - close DB connection
        """
        self.close()

    def close(self):
        """Close the cache's DB
        """
        if self.session is not None:
            self.session.close()
            self.session = None

    def _query(self, url, xpath):
        """Base query for an url and xpath

        Args:
            url (str): URL to search
            xpath (str): xpath to search (may be ``None``)
        """
        return self.session.query(CachedRequest).filter(CachedRequest.url == url).filter(CachedRequest.xpath == xpath)

    def get(self, url, store_on_error=False, xpath=None, rate_limit=None):
        """Get a URL via the cache.

        If the URL exists in the cache, return the cached value. Otherwise perform the request,
        store the resulting content in the cache and return it.

        Throws a :class:`RuntimeError` if the request results in an error.

        Args:
            url (str): URL to request
            store_on_error (bool): If True, store request results in cache even if request results in an
                an error. Otherwise (default) do not store results when an error occurs. Cached content
                equals exception message.
            xpath (str): If given (default is None), parses the response content to html, searches the first
                node matching the given xpath and returns only that node (as UTF8-encoded html). Also, only
                stores this node's html in the cache. Raises a ``RuntimeError`` if the xpath cannot be found
                in the response.
            rate_limit (float): If not None (default), wait at least this many seconds between the previous
                request and the current one (this does not apply to cache hits).

        Returns:
            str: request content
        """

        try:
            # get cached request - if none is found, this throws a NoResultFound exception
            cached = self._query(url, xpath).one()
            config.logger.info("Request cache hit: " + url)

            # if the cached value is from a request that resulted in an error, throw an exception
            if cached.status_code != requests.codes.ok:
                raise RuntimeError("Cached request returned an error, code " + str(cached.status_code))
        except NoResultFound:
            config.logger.info("Request cache miss: " + url)

            # perform the request
            try:

                # rate limit
                if rate_limit is not None and self.last_query is not None:
                    to_sleep = rate_limit - (datetime.datetime.now() - self.last_query).total_seconds()
                    if to_sleep > 0:
                        time.sleep(to_sleep)

                self.last_query = datetime.datetime.now()

                response = requests.get(url)
                status_code = response.status_code
                # get 'text', not 'content', because then we are sure to get unicode
                content = response.text
                response.close()

                if xpath is not None:
                    doc = html.fromstring(content)
                    nodes = doc.xpath(xpath)
                    if len(nodes) == 0:
                        # xpath not found; set content and status code, exception is raised below
                        content = "xpath not found: " + xpath
                        status_code = ERROR_XPATH_NOT_FOUND
                    else:
                        # extract desired node only
                        content = html.tostring(nodes[0], encoding='unicode')

            except requests.ConnectionError as e:
                # on a connection error, write exception information to a response object
                status_code = ERROR_CONNECTION_ERROR
                content = str(e)

            # a new request cache object
            cached = CachedRequest(url=str(url),
                                   content=content,
                                   status_code=status_code,
                                   xpath=xpath,
                                   queried_on=datetime.datetime.now())

            # if desired, store the response even if an error occurred
            if status_code == requests.codes.ok or store_on_error:
                self.session.add(cached)
                self.session.commit()

            if status_code != requests.codes.ok:
                raise RuntimeError("Error processing the request, " + str(status_code) + ": " + content)

        return cached.content

    def clear(self, url=None, xpath=None):
        """Clear cache

        Args:
            url (str): If given, clear specific item only. Otherwise remove the DB file.
            xpath (str): xpath to search (may be ``None``)
        """
        if url is not None:
            query = self._query(url, xpath)
            if query.count() > 0:
                query.delete()
                self.session.commit()
            else:
                raise KeyError("Cannot clear URL, not in cache: " + str(url) + " xpath:" + str(xpath))
        else:
            # remove the DB file
            self.close()
            if path.exists(self.db_path):
                remove(self.db_path)

    def has(self, url, xpath=None):
        """Check if a URL (and xpath) exists in the cache

        If DB has not been initialized yet, returns ``False`` for any URL.

        Args:
            url (str): If given, clear specific item only. Otherwise remove the DB file.
            xpath (str): xpath to search (may be ``None``)

        Returns:
            bool: ``True`` if URL exists, ``False`` otherwise
        """
        if not path.exists(self.db_path):
            return False

        return self._query(url, xpath).count() > 0

    def get_timestamp(self, url, xpath=None):
        """Get time stamp of cached query result.

        If DB has not yet been initialized or url/xpath has not been queried yet, return None.

        Args:
            url (str): If given, clear specific item only. Otherwise remove the DB file.
            xpath (str): xpath to search (may be ``None``)

        Returns:
            datetime.datetime: cached response timestamp, None if not available
        """
        if not path.exists(self.db_path):
            return None

        if self._query(url, xpath).count() > 0:
            return self._query(url, xpath).one().queried_on



"""ozelot configuration

This module defines default configuration values for the ozelot package.
If you want to change any of these values, create a file ``project_config.py``
and place it anywhere in the Python path used for the respective project.
If such a module is found, any variables found in it will be added to the configuration.
Default values will be stored as ``<variable name>_DEFAULT``.

At the very least, your ``project_config.py`` will have to define the connection
parameters for the database you want to use, and set these as ``DEFAULT_DB_PARAMS``.

Database connection parameters are stored as dictionaries with the following structure:

.. code-block:: python

    TEMPLATE_DB_PARAMS = {
        'driver': 'postgresql+psycopg2',
        'host': 'my.server.com',
        'port': 5432,
        'user': 'myuser',
        'password': None,
        'database': 'mydb'
    }

Multiple sets of connection parameters can be stored, as separate variables (with names of your choosing).
The pipeline will use the ``DEFAULT_DB_PARAMS`` by default. If needed, you can switch between different
configurations by setting this parameter to your desired parameter set, before starting any pipeline operations.

The ``driver`` is the driver + dialect string for :mod:`sqlalchemy`. See
`the documentation <http://docs.sqlalchemy.org/en/latest/core/engines.html#supported-databases>`_ for details
on supported databases and drivers.
Pass the desired dictionary into :class:`ozelot.client.Client` to create a database connection.
``user``, ``password``, ``host`` and ``port`` can be omitted (or set to ``None``) if not needed.
For SQLite file-based databases, set ``database`` to the file path and omit ``host``, ``port``,
``user`` and ``password``.

If the password field is omitted from the dictionary (or set to ``None``), :class:`ozelot.client.Client` will
try to retrieve it from a :mod:`keyring` password store, using the ``user`` field as the user name for
keyring and ``<host>:<driver>`` as service name. :meth:`ozelot.client.Client.store_password` is a convenience function
for storing the password in keyring. Set ``password`` to an empty string to not use a password.

If you choose to store the password as clear text, make sure to set permissions appropriately
(``0600``) on the ``project_config.py`` file.

Attributes:
    TESTING_SQLITE_DB_PATH (str): Path to use for the SQLite database file used in unit tests.
        Default: ``testing.db`` in the directory of the tests package.

    LOG_LEVEL (int): Default log level for :mod:logging

    TEMPLATE_DB_PARAMS (dict): Template dictionary for definition of database connection parameters.

    DEFAULT_DB_PARAMS (dict): Default database parameters to use.

    REQUEST_CACHE_PATH: File path where to store request cache (note: '.db' will be appended to the given path)

"""
from builtins import str

import sys
from os import path
import logging
import types

#
# Logging
#

# configure a basic logging client
logger = logging.getLogger('ozelot')
logger_handler = logging.StreamHandler()
logger_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger_handler.setFormatter(logger_formatter)
logger.addHandler(logger_handler)

# the default log level
LOG_LEVEL = logging.INFO

# the log level may have been overwritten in the project-specific config, need
# to retrieve it here, because loading of the project-specific config already logs
try:
    # noinspection PyUnresolvedReferences
    import project_config

    logger.setLevel(getattr(project_config, 'LOG_LEVEL', LOG_LEVEL))
except ImportError:
    logger.setLevel(LOG_LEVEL)
logger_handler.setLevel(logger.level)

logger.debug("Configured logging!")

#
# Databases
#

# The path to the SQLite DB file to use for unit tests
try:
    # noinspection PyUnresolvedReferences
    from ozelot import tests

    TESTING_SQLITE_DB_PATH = path.join(path.dirname(tests.__file__), 'testing.db')
except ImportError:
    # tests not present, ignore
    pass

# A template for defining database connections (see docstring at top for explanation)
TEMPLATE_DB_PARAMS = {
    'driver': 'postgresql+psycopg2',
    'host': 'my.server.com',
    'port': 5432,
    'user': 'myuser',
    'password': None,
    'database': 'mydb'
}

DEFAULT_DB_PARAMS = TEMPLATE_DB_PARAMS

#
# Web request cache
#

# The path to the file to store the request cache in - default is next to this file as '_web_cache.db'

REQUEST_CACHE_PATH = path.join(path.dirname(__file__), '_web_cache.db')

#
# Supporting tools
#

# Location of 'dot' executable, required (only) for rendering data model and pipeline diagrams.
# Default: in same directory as the Python interpreter
DOT_EXECUTABLE = path.join(path.dirname(sys.executable), 'dot')

#
# Load the project-specific configuration
#

# try importing project-specific configuration
try:
    # noinspection PyUnresolvedReferences
    import project_config

    logger.info("Reading project configuration from " + project_config.__file__)

    # get reference to current module
    this_module = sys.modules[__name__]

    # import all non-private, non-module attributes starting with an uppercase letter
    for attr in dir(project_config):
        if not attr.startswith('__'):

            old_value = getattr(this_module, attr, None)

            # skip modules
            if isinstance(old_value, types.ModuleType):
                continue

            # make backup of old value
            if old_value is not None:
                setattr(this_module, attr + '_DEFAULT', getattr(this_module, attr))

            # set new value
            new_value = getattr(project_config, attr)
            setattr(this_module, attr, new_value)

            logger.debug(u"\tRead configuration value from project_config: {:s} = {:s}"
                         u"".format(attr, str(new_value)))

except ImportError:
    logger.debug("Could not find a 'project_config' module to import")



"""Extensions of :mod:`luigi.Task`
"""
from builtins import str
from builtins import object
import luigi

from ozelot.etl.targets import ORMTarget
from ozelot import client
from ozelot import config


def get_task_name(task):
    """Return the task name (= class name)
    """
    return task.__class__.__name__


def get_task_param_string(task):
    """Get all parameters of a task as one string

    Returns:
        str: task parameter string
    """
    # get dict str -> str from luigi
    param_dict = task.to_str_params()

    # sort keys, serialize
    items = []
    for key in sorted(param_dict.keys()):
        items.append("'{:s}': '{:s}'".format(key, param_dict[key]))

    return "{" + ", ".join(items) + "}"


class TaskBase(luigi.Task):
    """Common functionality for :mod:`ozelot` tasks.
    """

    def clear(self):
        """Placeholder for clear method.

        This method should clear all output generated by the task, bringing the content database into the
        same state as just before running the task. This is not always easily possible. It is up to the
        developer to define and use clearing functions in such a way that her requirements are met.
        """
        raise NotImplementedError("Function 'clear' not implemented for this task.")

    @classmethod
    def get_name(cls):
        """Return the task name (= class name)
        """
        return cls.__name__

    def get_param_string(self):
        """Get all parameters of a task as one string

        Returns:
            str: task parameter string
        """
        return get_task_param_string(self)


class ORMTask(TaskBase):
    """A task with one target of type :class:`ozelot.models.target.ORMTarget` representing the task and its parameters.

    The default target type can be overwritten by setting the attribute :attr:`target_class`.
    """

    # node styling options for graph diagram
    diagram_style = ["bold"]

    # the target type to use
    target_class = ORMTarget

    def __init__(self, *args, **kwargs):
        """Initialize class attributes
        """
        super(ORMTask, self).__init__(*args, **kwargs)

        # DB session and client
        self._session = None
        self._client = None

    def clear(self):
        raise NotImplementedError("Function 'clear' not implemented for this task.")

    def output(self):
        """Task output: one target of type :class:`ozelot.models.target.ORMTarget` (or whatever else is defined
        as :attr:`target_class`).
        """
        yield self.target_class.from_task(self)

    def mark_complete(self):
        """Mark a task as complete - create all its target markers (by default there is just one).
        """
        for target in self.output():
            target.create()

    def mark_incomplete(self):
        """Mark a task as complete - create all its target markers (by default there is just one).

        Ignore if the task is not complete yet, i.e. a target does not exist.
        """
        for target in self.output():
            if target.exists():
                target.remove()

    @property
    def client(self):
        """Get the DB client associated with the task (open a new one if necessary)

        Returns:
            ozelot.client.Client: DB client
        """
        if self._client is None:
            self._client = client.get_client()

        return self._client

    @property
    def session(self):
        """Get the DB session associated with the task (open a new one if necessary)

        Returns:
            sqlalchemy.orm.session.Session: DB session
        """
        if self._session is None:
            self._session = self.client.create_session()

        return self._session

    def close_session(self, commit=True):
        """Commit and close the DB session associated with this task (no error is raised if None is open)

        Args:
            commit (bool): commit session before closing (default=True)
        """
        if self._session is not None:
            if commit:
                self._session.commit()
            self._session.close()
            self._session = None

    def done(self, commit=True):
        """Task cleanup: mark task as complete, commit and close the session (if open)

        Args:
            commit (bool): commit session before closing (default=True)
        """
        self.close_session(commit)
        self.mark_complete()


class ORMTestTask(ORMTask):
    """A testing task base class
    """

    # node styling options for graph diagram
    diagram_style = ["diagonals"]

    def clear(self):
        """Clearing the task means marking it incomplete (task stores no data)
        """
        self.mark_incomplete()

    def done(self, commit=False):
        """Override :func:`done` of :class:`ORMTask`: don't commit by default

        Args:
            commit (bool): commit session before closing (default=False)
        """
        super(ORMTestTask, self).done(commit)


# helper function: increment a stats counter
def _increment_stats(stats, label):
    count = stats.setdefault(label, 0)
    stats[label] = count + 1


def check_completion(task, mark_incomplete=False, clear=False, return_stats=False):
    """Recursively check if a task and all its requirements are complete

    Args:
        task (derived from luigi.Task): Task to check completion for; check everything 'downstream'
            from that task.

        mark_incomplete (bool): If ``True`` set any task as incomplete for which a requirement
            is found to be incomplete (checked recursively).
            This works only for tasks derived from :class:`ORMTask`.

        clear (bool): If ``True``, call the :func:`clear()` method of any task for which a requirement
            is found to be incomplete (checked recursively). This implies ``mark_incomplete = True``.
            This works only for tasks derived from :class:`ORMTask`.

        return_stats (bool): If ``True``, return task checking statistics in addition to completion status

    Returns:
        bool: ``True`` if the task, all its requirements and (recursively) all their requirements
            are complete, ``False`` otherwise.
    """
    # run recursive task checking, get stats
    to_clear = dict()
    is_complete, stats = _check_completion(task,
                                           mark_incomplete=mark_incomplete,
                                           clear=clear,
                                           stats={},
                                           visited=dict(),
                                           to_clear=to_clear)

    # task clearing needs to happen top-down: because of foreign key constraints, a task can
    # only be cleared once all tasks that require it have been cleared
    while to_clear:
        # find all tasks that we can currently clear - tasks not required by other tasks;
        # iterate over list of keys to be able to modify dict while iterating
        found_clearable_task = False
        for task_id in list(to_clear.keys()):
            v = to_clear[task_id]
            if not v['required_by']:
                # this is a task that can be cleared - no other task requires it
                found_clearable_task = True

                task = v['task']
                if isinstance(task, ORMTask):
                    task.mark_incomplete()
                    task.clear()
                    _increment_stats(stats, 'Cleared')
                    config.logger.info("Cleared task: " + task_id)
                else:
                    config.logger.info('Cannot clear task, not an ORMTask: ' + task_id)

                # remove the task from the list of tasks that need clearing, remove references
                # in the required_by lists of all other tasks; this is not an efficient implementation,
                # O(n^2), could be made O(n) using lookup tables of the task graph
                del to_clear[task_id]
                for w in to_clear.values():
                    w['required_by'].discard(task_id)

        if not found_clearable_task:
            raise RuntimeError("Error in recursive task clearing, no clearable task found")

    config.logger.info("Task completion checking, summary:\n" + str(stats))

    if return_stats:
        return is_complete, stats
    else:
        return is_complete


def _check_completion(task, mark_incomplete, clear, stats, visited, to_clear):
    """Core recursion function for :func:`check_completion`, see there for more documentation

    Args:
        task (luigi.Task): task instance

        mark_incomplete (bool): see :func:`check_completion`

        clear (bool): see :func:`check_completion`

        stats (dict): task checking statistics (counts of incomplete, to be cleared, ...)

        visited (dict): cache for visited tasks: key = task name + parameter string, value = completion status

        to_clear (dict): dict of dicts of tasks to be cleared, key = task id,
            value =  {task: task instance object, required_by: set of task IDs that this task is required by}

    """
    # is this task (recursively) complete?
    task_complete = task.complete()
    is_complete = task_complete

    # task identification: task name plus parameters
    task_id = get_task_name(task) + ' ' + get_task_param_string(task)

    # check any requirements
    for req in task.requires():

        # task/parameter ID string to identify this task instance
        req_id = get_task_name(req) + ' ' + get_task_param_string(req)

        # skip recursion on already visited tasks, get completion status from cache
        if req_id in visited:
            req_complete = visited[req_id]
        else:
            req_complete, _ = _check_completion(task=req,
                                                mark_incomplete=mark_incomplete,
                                                clear=clear,
                                                stats=stats,
                                                visited=visited,
                                                to_clear=to_clear)
            visited[req_id] = req_complete

        # add any incomplete requirements to the list of tasks to clear, noting the current task as parent (required by)
        if clear and not req_complete:
            clear_entry = to_clear.setdefault(req_id, dict(task=req, required_by=set()))
            clear_entry['required_by'].add(task_id)

        is_complete &= req_complete

    if not is_complete:
        if task_complete:
            config.logger.info("Task complete but requirements incomplete: " + task_id)
        else:
            config.logger.info("Task incomplete: " + task_id)
        _increment_stats(stats, 'Incomplete tasks')

        if mark_incomplete:
            if isinstance(task, ORMTask):
                task.mark_incomplete()
                _increment_stats(stats, 'Marked incomplete')
                config.logger.info("Marked task incomplete: " + task_id)
            else:
                config.logger.info('Cannot mark task incomplete, not an ORMTask: ' + task_id)

    else:
        _increment_stats(stats, 'Complete tasks')
        config.logger.debug("Task complete: " + task_id)

    # if we want to clear and the current task is not in the dict of tasks to clear,
    # it is the root task, add it with no parent (required by) tasks
    if clear and not is_complete and task_id not in to_clear:
        to_clear[task_id] = dict(task=task, required_by=set())

    return is_complete, stats


class ORMObjectCreatorMixin(object):
    """Functionality for tasks that are responsible for creating a certain class of ORM objects

    If a class (or multiple classes) of ORM objects are created only by a single task, clearing
    of the task output is simplified, as is bulk storing of task output to the database.
    """

    #: *list(ORM object classes)*: ORM object classes that are created by this task
    object_classes = list()

    def clear(
            self  # type: ORMTask
    ):
        """Delete all objects created by this task.

        Iterate over `self.object_classes` and delete all objects of the listed classes.
        """
        # mark this task as incomplete
        self.mark_incomplete()

        # delete objects
        for object_class in self.object_classes:
            self.session.query(object_class).delete()

        self.close_session()


class ORMWrapperTask(ORMTask):
    """A wrapper task that bundles other tasks and does not run its own code

    The task's :func:`run` method marks the task as complete and does nothing else.
    Since its requirements are checked before marking the task complete, it serves as
    a wrapper around other tasks, for logical grouping.

    The tasks :func:`complete` method only returns complete if the task itself
    is marked complete and all requirements are complete. This behaviour is useful
    when adding tasks to a wrapper task during development -- adding a new tasks
    triggers re-running the wrapper task.
    """

    # node styling options for graph diagram
    diagram_style = ["dotted"]

    def clear(self):
        """Mark task as incomplete, do nothing else.
        """
        self.mark_incomplete()

    def run(self):
        """Mark task as complete, do nothing else.
        """
        self.mark_complete()

    def complete(self):
        """Task is complete if completion marker is set and all requirements are complete
        """
        is_complete = super(ORMWrapperTask, self).complete()
        for req in self.requires():
            is_complete &= req.complete()

        return is_complete


class InputFileTask(TaskBase):
    """A base class for tasks representing input files.

    This task has an empty run method (does nothing) and specifies the input file as task output.
    Existence of the input file signals task completion; this is checked by , :mod:`luigi`.

    It is useful to supply a `load` method in the derived classes, which loads the respective data file
    to a usable format (e.g., to a :class:`pandas.DataFrame`).
    """

    # the path to the input file, to be defined in the derived class
    input_file = None

    # node styling options for graph diagram
    diagram_style = ["rounded"]

    def output(self):
        """Task output: one :class:`luigi.LocalTarget` for the input file
        """
        if self.input_file is None:
            raise RuntimeError("Please specify 'input_file' for the task class")

        return [luigi.LocalTarget(self.input_file)]

    def clear(self):
        """Input files should never be cleared - log the call but do nothing else.
        """
        config.logger.info("Called 'clear' for input file task {:s} - doing nothing.".format(self.get_name()))

    def run(self):
        """Run method - does nothing
        """
        pass

    def load(self):
        """A method to load the input file to a usable format, to be defined in the derived class.
        """
        raise NotImplementedError()



"""Target classes for :mod:`luigi`."""
from builtins import str

import datetime

from sqlalchemy import Column, String, DateTime

from ozelot.orm import base


class ORMTargetMarker(base.Base):
    # noinspection PyUnresolvedReferences
    """ORM object class to store task completion markers.
    """

    #: *Column(String)*: target name, usually equals name of the associated task
    name = Column(String())

    #: *Column(String)*: string representation of task parameters
    params = Column(String())

    #: *Column(DateTime)*: Time stamp for object creation
    created = Column(DateTime, default=datetime.datetime.utcnow)

    def __str__(self):
        return self.name + " " + self.params + " " + str(self.created)



import os


"""
Utility functions for ETL pipelines
"""


def render_diagram(root_task, out_base, max_param_len=20, horizontal=False):
    """Render a diagram of the ETL pipeline

    All upstream tasks (i.e. requirements) of :attr:`root_task` are rendered.

    Nodes are, by default, styled as simple rects. This style is augmented by any
    :attr:`diagram_style` attributes of the tasks.

    .. note:: This function requires the 'dot' executable from the GraphViz package to be installed
              and its location configured in your `project_config.py` variable :attr:`DOT_EXECUTABLE`.

    Args:
        root_task (luigi.Task): Task instance that defines the 'upstream root' of the pipeline
        out_base (str): base output file name (file endings will be appended)
        max_param_len (int): Maximum shown length of task parameter values
        horizontal (bool): If True, layout graph left-to-right instead of top-to-bottom
    """
    import re
    import codecs
    import subprocess
    from ozelot import config
    from ozelot.etl.tasks import get_task_name, get_task_param_string

    # the graph - lines in dot file
    lines = [u"digraph G {"]

    if horizontal:
        lines.append(u"rankdir=LR;")

    # helper function: make unique task id from task name and parameters:
    # task name + parameter string, with spaces replaced with _ and all non-alphanumerical characters stripped
    def get_id(task):
        s = get_task_name(task) + "_" + get_task_param_string(task)
        return re.sub(r'\W+', '', re.sub(' ', '_', s))

    # node names of tasks that have already been added to the graph
    existing_nodes = set()

    # edge sets (tuples of two node names) that have already been added
    existing_edges = set()

    # recursion function for generating the pipeline graph
    def _build(task, parent_id=None):
        tid = get_id(task)

        # add node if it's not already there
        if tid not in existing_nodes:
            # build task label: task name plus dictionary of parameters as table

            params = task.to_str_params()
            param_list = ""
            for k, v in params.items():
                # truncate param value if necessary, and add "..."
                if len(v) > max_param_len:
                    v = v[:max_param_len] + "..."
                param_list += "<TR><TD ALIGN=\"LEFT\">" \
                              "<FONT POINT-SIZE=\"10\">{:s}</FONT>" \
                              "</TD><TD ALIGN=\"LEFT\">" \
                              "<FONT POINT-SIZE=\"10\">{:s}</FONT>" \
                              "</TD></TR>".format(k, v)

            label = "<TABLE BORDER=\"0\" CELLSPACING=\"1\" CELLPADDING=\"1\">" \
                    "<TR><TD COLSPAN=\"2\" ALIGN=\"CENTER\">" \
                    "<FONT POINT-SIZE=\"12\">{:s}</FONT>" \
                    "</TD></TR>" \
                    "".format(get_task_name(task)) + param_list + "</TABLE>"

            # add a node for the task
            lines.append(u"{name:s} [label=< {label:s} >, shape=\"rect\", style=\"{style:s}\"];\n"
                         u"".format(name=tid,
                                    label=label,
                                    style=','.join(getattr(task, 'diagram_style', []))))

            existing_nodes.add(tid)

            # recurse over requirements
            for req in task.requires():
                _build(req, parent_id=tid)

        # add edge from current node to (upstream) parent, if it doesn't already exist
        if parent_id is not None and (tid, parent_id) not in existing_edges:
            lines.append(u"{source:s} -> {target:s};\n".format(source=tid, target=parent_id))

    # generate pipeline graph
    _build(root_task)

    # close the graph definition
    lines.append(u"}")

    # write description in DOT format
    with codecs.open(out_base + '.dot', 'w', encoding='utf-8') as f:
        f.write(u"\n".join(lines))

    # check existence of DOT_EXECUTABLE variable and file
    if not hasattr(config, 'DOT_EXECUTABLE'):
        raise RuntimeError("Please configure the 'DOT_EXECUTABLE' variable in your 'project_config.py'")
    if not os.path.exists(config.DOT_EXECUTABLE):
        raise IOError("Could not find file pointed to by 'DOT_EXECUTABLE': " + str(config.DOT_EXECUTABLE))

    # render to image using DOT
    # noinspection PyUnresolvedReferences
    subprocess.check_call([
        config.DOT_EXECUTABLE,
        '-T', 'png',
        '-o', out_base + '.png',
              out_base + '.dot'
    ])



"""Database client and singleton client instance
"""
from builtins import str
from builtins import object

import keyring
import sqlalchemy as sa
from sqlalchemy import orm

from ozelot import config


# the singleton client instance
_client = None


def get_client():
    """Get or create the singleton client instance
    """
    global _client

    if _client is None:
        _client = Client(params=config.DEFAULT_DB_PARAMS)

    return _client


def set_client(client):
    """Store a client instance as singleton client

    Args:
        client (ozelot.client.Client): database client
    """
    global _client

    _client = client


class Client(object):
    """The database client.
    """

    def __init__(self, params=None, connection_string=None):
        """Instantiate a client object

        A client can be configured either from a parameters dictionary ``params`` or directly
        from an :mod:`sqlalchemy` connection string ``connection_string``. Exactly one of the two
        must be provided.

        Args:
            params (dict): database configuration, as defined in :mod:`ozelot.config`
            connection_string (str): :mod:`sqlalchemy` connection string
        """

        if params is None and connection_string is None:
            raise RuntimeError("Please provide either 'params' or 'connection_string'")

        if params is not None and connection_string is not None:
            raise RuntimeError("Please provide only on of 'params' or 'connection_string'")

        if params is not None:
            # log connection string with password hidden
            # noinspection PyTypeChecker
            connection_string_no_pw = self.get_connection_string(params=params, hide_password=True)
            config.logger.info("Client connecting to: " + connection_string_no_pw)

            # noinspection PyTypeChecker
            connection_string = self.get_connection_string(params=params, hide_password=False)

        else:
            # noinspection PyTypeChecker
            config.logger.info("Client connecting to: " + connection_string)

        # create the engine
        self.engine = sa.create_engine(connection_string)

        # turn on foreign key support for SQLite (required for cascading deletes etc.)
        if connection_string.startswith('sqlite://'):

            def on_connect(conn, _):
                conn.execute('pragma foreign_keys=ON')

            from sqlalchemy import event
            event.listen(self.engine, 'connect', on_connect)

        # create a session factory
        self.session_maker = orm.sessionmaker(bind=self.get_engine())

    def get_engine(self):
        """Get the database engine
        """
        return self.engine

    def create_session(self):
        """Create a new :mod:`sqlalchemy` ORM session

        .. seealso:: `sqlalchemy documentation on sessions <http://docs.sqlalchemy.org/en/latest/orm/session.html>`_
        """
        return self.session_maker()

    @staticmethod
    def store_password(params, password):
        """Store the password for a database connection using :mod:`keyring`

        Use the ``user`` field as the user name and ``<host>:<driver>`` as service name.

        Args:
            params (dict): database configuration, as defined in :mod:`ozelot.config`
            password (str): password to store
        """
        user_name = params['user']
        service_name = params['host'] + ':' + params['driver']
        keyring.set_password(service_name=service_name,
                             username=user_name,
                             password=password)

    @staticmethod
    def _get_password(params):
        """Get the password for a database connection from :mod:`keyring`

        Args:
            params (dict): database configuration, as defined in :mod:`ozelot.config`

        Returns:
            str: password
        """
        user_name = params['user']
        service_name = params['host'] + ':' + params['driver']
        return keyring.get_password(service_name=service_name,
                                    username=user_name)

    @staticmethod
    def get_connection_string(params, hide_password=True):
        """Get a database connection string

        Args:
            params (dict): database configuration, as defined in :mod:`ozelot.config`
            hide_password (bool): if True, the password is hidden in the returned string
                (use this for logging purposes).

        Returns:
            str: connection string
        """
        connection_string = params['driver'] + '://'

        user = params.get('user', None)
        password = params.get('password', None)
        host = params.get('host', None)
        port = params.get('port', None)
        database = params.get('database', None)

        if database is None:
            raise ValueError("Field 'database' of connection parameters cannot be None.")

        # if password is not set, try to get it from keyring
        if password is None and user is not None:
            # noinspection PyTypeChecker
            password = Client._get_password(params)

            if password is None:
                raise RuntimeError("Password not defined and not available in keyring.")

        # don't add host/port/user/password if no host given
        if host is not None:

            # don't add user/password if user not given
            if user is not None:
                connection_string += user

                # omit zero-length passwords
                if len(password) > 0:
                    if hide_password:
                        connection_string += ":[password hidden]"
                    else:
                        connection_string += ":" + password

                connection_string += "@"

            connection_string += host

            if port is not None:
                connection_string += ':' + str(port)

        # noinspection PyTypeChecker
        connection_string += '/' + database

        return connection_string

    def df_query(self, query, with_labels=False):
        """
        Run a :mod:`sqlalchemy` query and return result as a :class:`pandas.DataFrame`

        Args:

            query (sqlalchemy.orm.query.Query): query object, usually generated by :func:`session.query()` in
                an :class:`sqlalchemy.orm.session.Session`

            with_labels (bool): A query for fields with the same name from different tables will cause problems
                when converting it to a :class:`pandas.DataFrame`, because there will be duplicate column names.
                When setting `with_labels=True`, disambiguation labels are assigned to all (!)
                fields in the query - the field name is prefixed with the column name. This enables
                querying fields with identical names from multiple tables but getting unique column names in the output.

        :return: query result as :class:`pandas.DataFrame`
        """
        import pandas as pd

        if with_labels:
            query = query.with_labels()

        # compile sql statement, including arguments
        statement = query.statement.compile(self.engine)

        # run query
        return pd.read_sql_query(sql=statement, con=self.engine)




from builtins import object
from ozelot import client
from ozelot.orm.target import ORMTargetMarker


class ORMTarget(object):
    """Pipeline target that stores task completion markers in a database, via an ORM layer.
    """

    def __init__(self, name, params):
        """Create a new target instance

        Args:
            name: task name
            params: task parameters
        """
        self.name = name
        self.params = params

    @classmethod
    def from_task(cls, task):
        """Create a new target representing a task and its parameters

        Args:
            task: Task instance to create target for; the task class has to inherit
                from :class:`ozelot.tasks.TaskBase`.

        Returns:
            ozelot.tasks.ORMTarget: a new target instance
        """
        target = cls(name=task.get_name(),
                     params=task.get_param_string())

        return target

    def _base_query(self, session):
        """Base query for a target.

        Args:
            session: database session to query in
        """
        return session.query(ORMTargetMarker) \
            .filter(ORMTargetMarker.name == self.name) \
            .filter(ORMTargetMarker.params == self.params)

    def exists(self):
        """Check if a target exists

        This function is called by :mod:`luigi` to check if a task output exists. By default,
        :mod:`luigi` considers a task as complete if all it targets (outputs) exist.

        Returns:
            bool: ``True`` if target exists, ``False`` otherwise
        """
        # get DB connection
        session = client.get_client().create_session()

        # query for target existence
        ret = self._base_query(session).count() > 0

        session.close()

        return ret

    def create(self):
        """Create an instance of the current target in the database

        If a target with the current name and params already exists, no second instance is created.
        """
        session = client.get_client().create_session()

        if not self._base_query(session).count() > 0:
            # store a new target instance to the database
            marker = ORMTargetMarker(name=self.name,
                                     params=self.params)
            session.add(marker)
            session.commit()

        session.close()

    def remove(self):
        """Remove a target

        Raises a ``RuntimeError`` if the target does not exist.
        """
        session = client.get_client().create_session()

        if not self._base_query(session).count() > 0:
            session.close()
            raise RuntimeError("Target does not exist, name={:s}, params={:s}"
                               "".format(self.name, self.params))

        # remove the target from the database
        self._base_query(session).delete()
        session.commit()

        session.close()



from setuptools import setup

setup(name='ozelot',
      version='0.2.2',
      description='A package for building maintanable ETL pipelines',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Topic :: Scientific/Engineering :: Information Analysis',
      ],
      keywords='data analytics ETL',
      url='https://github.com/trycs/ozelot',
      author='Eberhard Hansis',
      author_email='contact@transparent-analytics.de',
      license='MIT',
      packages=['ozelot',
                'ozelot.orm',
                'ozelot.etl'],
      install_requires=[
          'pandas>=0.17.0',
          'sqlalchemy>=1.0.0',
          'luigi>=2.2.0',
          'lxml',
          'keyring',
          'requests',
          'sadisplay',
          'future',
      ],
      test_suite='nose.collector',
      tests_require=['nose',
                     'coverage',
                     'lxml'],
      include_package_data=True,
      zip_safe=False)



"""This module defines and instantiates an extended declarative base class for ORM objects.

Attributes:

    Base (ExtendedBase): Instance of extended declarative base class; derive your models from this.

    model_registry (dict): Registry of all model classes derived from the base, key = class name, value = class

Define derived classes like this::

    from sqlalchemy import Column, Integer
    from ozelot.models import base

    class MyModel(base.Base):
        my_field = Column(Integer())
"""
import os
from builtins import object

from sqlalchemy import Column, Integer, func
from sqlalchemy.ext.declarative import declared_attr, DeclarativeMeta
from sqlalchemy.ext.declarative import declarative_base

from ozelot import config

model_registry = {}


class RegistryMeta(DeclarativeMeta):
    """Metaclass that registers all objects derived from the base class with the model registry

    Exclude the base class instance itself from the registration.
    """

    # noinspection PyInitNewSignature
    def __init__(cls, name, bases, class_dict):
        if cls.__name__ not in model_registry and cls.__name__ != 'Base':
            model_registry[cls.__name__] = cls
        super(RegistryMeta, cls).__init__(name, bases, class_dict)


class ExtendedBase(object):
    # noinspection PyUnresolvedReferences
    """Extension for :mod:`sqlalchemy` ORM base class.

        Most methods of this class could be class methods. However, the :mod:`sqlalchemy`
        ``@declared_attr`` seems to not play nice with the ``@classmethod`` decorator.
        Methods like :meth`create` or :meth:`drop` have to be called with instance of the
        respective derived model class.

        Attributes:
            __tablename__ (str): Default table name equalling the lower-cased class name
        """

    #: Default integer primary key column
    id = Column(Integer, primary_key=True)

    @declared_attr
    def __tablename__(self):
        return self.__name__.lower()

    def create_table(self, client):
        """Create the database table belonging to the model.

        If the table already exists, :mod:`sqlalchemy` will throw an error.

        Args:
            client (ozelot.client.Client): database client to use
        """
        # noinspection PyUnresolvedReferences
        self.__table__.create(bind=client.get_engine())

    def drop_table(self, client):
        """Drop the database table belonging to the model

        This will fail if any table row id is still being referenced as a foreign key from another table,
        and if the database backend enforces foreign key constraints (SQLite does not).

        Args:
            client (ozelot.client.Client): database client to use
        """
        # noinspection PyUnresolvedReferences
        self.__table__.drop(bind=client.get_engine())

    @classmethod
    def create_all(cls, client):
        """Create all tables derived from this base class

        Before you call this, import all modules that define the models that you want to create
        tables for (this registers your models with the base class metadata object).

        Args:
            client (ozelot.client.Client): database client to use
        """
        # noinspection PyUnresolvedReferences
        cls.metadata.create_all(bind=client.get_engine())

    @classmethod
    def drop_all(cls, client):
        """Drop all tables derived from this base class

        Before you call this, import all modules that define the models that you want to drop
        tables for (this registers your models with the base class metadata object).

        Args:
            client (ozelot.client.Client): database client to use
        """
        # noinspection PyUnresolvedReferences
        cls.metadata.drop_all(bind=client.get_engine())

    @classmethod
    def get_max_id(cls, session):
        """Get the current max value of the ``id`` column.

        When creating and storing ORM objects in bulk, :mod:`sqlalchemy` does not automatically
        generate an incrementing primary key ``id``. To do this manually, one needs to know the
        current max ``id``. For ORM object classes that are derived from other ORM object classes,
        the max ``id`` of the lowest base class is returned. This is designed to be used with
        inheritance by joining, in which derived and base class objects have identical ``id`` values.

        Args:
            session: database session to operate in
        """
        # sqlalchemy allows only one level of inheritance, so just check this class and all its bases
        id_base = None
        for c in [cls] + list(cls.__bases__):
            for base_class in c.__bases__:
                if base_class.__name__ == 'Base':
                    if id_base is None:
                        # we found our base class for determining the ID
                        id_base = c
                    else:
                        raise RuntimeError("Multiple base object classes for class " + cls.__name__)

        # this should never happen
        if id_base is None:
            raise RuntimeError("Error searching for base class of " + cls.__name__)

        # get its max ID
        max_id = session.query(func.max(id_base.id)).scalar()

        # if no object is present, None is returned
        if max_id is None:
            max_id = 0

        return max_id

    def truncate_to_field_length(self, field, value):
        """Truncate the value of a string field to the field's max length.

        Use this in a validator to check/truncate values before inserting them into the database.
        Copy the below example code after ``@validates`` to your model class and replace ``field1`` and ``field2`` with
        your field name(s).

        :Example:

            from sqlalchemy.orm import validates
            # ... omitting other imports ...

            class MyModel(base.Base):

                field1 = Column(String(128))
                field2 = Column(String(64))

                @validates('field1', 'field2')
                def truncate(self, field, value):
                    return self.truncate_to_field_length(field, value)

        Args:
            field (str): field name to validate
            value (str/unicode): value to validate

        Returns:
            str/unicode: value truncated to field max length

        """
        max_len = getattr(self.__class__, field).prop.columns[0].type.length
        if value and len(value) > max_len:
            return value[:max_len]
        else:
            return value


Base = declarative_base(cls=ExtendedBase, metaclass=RegistryMeta)


def render_diagram(out_base):
    """Render a data model diagram

    Included in the diagram are all classes from the model registry.
    For your project, write a small script that imports all models that you would like to
    have included and then calls this function.

    .. note:: This function requires the 'dot' executable from the GraphViz package to be installed
              and its location configured in your `project_config.py` variable :attr:`DOT_EXECUTABLE`.

    Args:
        out_base (str): output base path (file endings will be appended)
    """
    import codecs
    import subprocess
    import sadisplay

    # generate class descriptions
    desc = sadisplay.describe(list(model_registry.values()),
                              show_methods=False,
                              show_properties=True,
                              show_indexes=True,
                              )

    # write description in DOT format
    with codecs.open(out_base + '.dot', 'w', encoding='utf-8') as f:
        f.write(sadisplay.dot(desc))

    # check existence of DOT_EXECUTABLE variable and file
    if not hasattr(config, 'DOT_EXECUTABLE'):
        raise RuntimeError("Please configure the 'DOT_EXECUTABLE' variable in your 'project_config.py'")
    if not os.path.exists(config.DOT_EXECUTABLE):
        raise IOError("Could not find file pointed to by 'DOT_EXECUTABLE': " + str(config.DOT_EXECUTABLE))

    # render to image using DOT
    # noinspection PyUnresolvedReferences
    subprocess.check_call([
        config.DOT_EXECUTABLE,
        '-T', 'png',
        '-o', out_base + '.png',
              out_base + '.dot'
    ])
